/*  display.c
 *
 *  Display/Rendering functions for xpsk31
 */

/*
 *  xpsk31: An application to transmit and receive
 *  PSK31 signals using the computer's sound card
 *
 *  Copyright (C) 2003  Neoklis Kyriazis
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details:
 *
 *  http://www.gnu.org/copyleft/gpl.txt
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gtk/gtk.h>
#include "xpsk31.h"
#include "support.h"

/* xpsk31 main window */
extern GtkWidget *main_window;

/* Global widgets */
extern GtkWidget
  *gbl_scope,
  *gbl_drawingarea,
  *gbl_tx_scrolledwindow;

/* Waterfall window pixbuf */
extern GdkPixbuf *gbl_wfall_pixbuf;
extern guchar *gbl_wfall_pixels;
extern gint
  gbl_wfall_rowstride,
  gbl_wfall_n_channels,
  gbl_wfall_width,
  gbl_wfall_height;

/* Text buffers for text viewers */
extern GtkTextBuffer *gbl_tx_text_buffer;

/* Index to macros */
int gbl_macro_idx;

/* Runtime config data */
extern rc_data_t rc_data;

/* Average bin values */
int gbl_bin_ave[FFT_SIZE/2];

/* fft in/out buffers */
extern int
  *fft_in_r,
  *fft_out_r,
  *fft_out_i;

/* Audio freq error */
extern int freq_error;

/* Operator and location data */
extern qso_record_t qso_record;

extern FILE
  /* File for recording QSO's */
  *qso_record_fp,
  /* File for ADIF QSO record */
  *log_adif_fp,
  /* File for station log */
  *log_fp,
  /* Pointer to samples file */
  *samples_fp;

/*------------------------------------------------------------------------*/

/* Display_Waterfall()
 *
 * Displays audio spectrum as "waterfall"
 */

  void
Display_Waterfall( void )
{
  int
	idh, idv,  /* Index to hor. and vert. position in warterfall */
	pixel_val, /* Greyscale value of pixel derived from fft o/p  */
	fft_idx;   /* Index to fft output array */

  /* Constant needed to draw red line in waterfall */
  static int red_line = WFALL_WIDTH/2;

  int
	bin_val, /* Value of fft output "bin" */
	bin_max, /* Maximum value of fft bins */
	temp;

  /* Sliding window average of bin max values */
  static int ave_max = 1000;

  /* Pointer to current pixel */
  static guchar *pix;

  /* Draw a vertical red line in waterfall at detector's freq. */
  pix = gbl_wfall_pixels + gbl_wfall_rowstride + gbl_wfall_n_channels;
  pix += gbl_wfall_n_channels * red_line;
  pix[0] = 0xff;
  pix[1] = pix[2] = 0;

  /* Copy each line of waterfall to next one */
  temp = gbl_wfall_height-2;
  for( idv = temp; idv > 0; idv-- )
  {
	pix = gbl_wfall_pixels + gbl_wfall_rowstride * idv + gbl_wfall_n_channels;

	for( idh = 0; idh < WFALL_WIDTH; idh++ )
	{
	  *pix = *( pix - gbl_wfall_rowstride ); pix++;
	  *pix = *( pix - gbl_wfall_rowstride ); pix++;
	  *pix = *( pix - gbl_wfall_rowstride );
	  pix += gbl_wfall_n_channels - 2;
	}
  }

  /* Got to top left +1 hor. +1 vert. of pixbuf */
  pix = gbl_wfall_pixels + gbl_wfall_rowstride + gbl_wfall_n_channels;

  /* First (DC) fft bin not used */
  gbl_bin_ave[0] = 0;

  /* Do the FFT on input array */
  Ifft();
  bin_max = 0;
  temp = (2*WFALL_WIDTH)/(FFT_SIZE);
  for( fft_idx = 0; fft_idx < FFT_SIZE2; fft_idx++ )
  {
	/* Calculate power of signal at each freq. ("bin") */
	fft_out_r[fft_idx] /= 256;
	fft_out_i[fft_idx] /= 256;
	bin_val =
	  fft_out_r[fft_idx]*fft_out_r[fft_idx] +
	  fft_out_i[fft_idx]*fft_out_i[fft_idx];

	/* Calculate average bin values */
	gbl_bin_ave[fft_idx] = ( (gbl_bin_ave[fft_idx] *
		  (BINMAX_AVE_WIN-1) + bin_val) / BINMAX_AVE_WIN );

	/* Record max bin value */
	if( bin_max < bin_val )
	  bin_max = bin_val;

	/* Scale pixel values to 255 depending on ave max value */
	pixel_val = (255 * bin_val) / ave_max;
	if( pixel_val > 255 )
	  pixel_val = 255;

	/* Duplicate pixels if fft size < waterfall width */
	for( idh = 0; idh < temp; idh++ )
	{
	  /* Reverse video is toggled by right click on waterfall */
	  if( isFlagSet(REVERSE_VIDEO) )
		pix[0] = pix[1] = pix[2] = pixel_val;
	  else
		pix[0] = pix[1] = pix[2] = 255 - pixel_val;

	  pix += gbl_wfall_n_channels;
	}

  } /* for( fft_idx = 0; fft_idx < FFT_SIZE2; fft_idx++ ) */

  /* Calculate sliding window average of max bin value */
  ave_max = (ave_max*(BINMAX_AVE_WIN-1) + bin_max)/BINMAX_AVE_WIN;
  if( !ave_max ) ave_max = 1;

  /* At last draw waterfall */
  gtk_widget_queue_draw( gbl_scope );

} /* Display_Waterfall() */

/*------------------------------------------------------------------------*/

/*  Clear_Wfall_Pixbuf()
 *
 *  Initializes pixbuf of Receive window
 */

  void
Clear_Wfall_Pixbuf( void )
{
  int i;
  guchar *p1, *p2;

  gdk_pixbuf_fill( gbl_wfall_pixbuf, 0xffffffff );

  /* Draw a box around pixbuf */
  p1 = gbl_wfall_pixels;
  p2 = gbl_wfall_pixels + (gbl_wfall_height-1)*gbl_wfall_rowstride;
  for( i = 0; i < gbl_wfall_width; i++ )
  {
	p1[0] = p1[1] = p1[2] = 0;
	p1 += gbl_wfall_n_channels;
	p2[0] = p2[1] = p2[2] = 0;
	p2 += gbl_wfall_n_channels;
  }

  p1 = gbl_wfall_pixels+gbl_wfall_rowstride;
  p2 = gbl_wfall_pixels+gbl_wfall_rowstride +
	(gbl_wfall_width-1)*gbl_wfall_n_channels;
  for( i = 0; i < gbl_wfall_height-1; i++ )
  {
	p1[0] = p1[1] = p1[2] = 0;
	p1 += gbl_wfall_rowstride;
	p2[0] = p2[1] = p2[2] = 0;
	p2 += gbl_wfall_rowstride;
  }

  gtk_widget_queue_draw( gbl_scope );

} /* Clear_Wfall_Pixbuf() */

/*------------------------------------------------------------------------*/

/*  Update_Magniphase()
 *
 *  Displays carrier phase and magnitude
 */

  void
Update_Magniphase( int level_i, int level_q )
{
  int x, y;

  /* Scale factors needed to draw freq error pointer */
  static int
	sx = 3*32*32,
    sy = 32;

  /* Draw a white filled circle */
  gdk_draw_arc( gbl_scope->window,
	  gbl_scope->style->white_gc, TRUE,
	  0, 0, 128, 128, 0, 23040 );

  /* Draw a black outline around above */
  gdk_draw_arc( gbl_scope->window,
	  gbl_scope->style->black_gc, FALSE,
	  0, 0, 128, 128, 0, 23040 );

  /* Draw the gbl_scope vector */
  gdk_draw_line( gbl_scope->window,
	  gbl_scope->style->black_gc,
	  64, 64, 64+level_q, 64-level_i );

  /* Draw freq error pointer */
  x = freq_error*(2*sx-freq_error*freq_error)/sx;
  y = 64 - (2*sy*sy-freq_error*freq_error)/sy;
  gdk_draw_line( gbl_scope->window,
	  gbl_scope->style->black_gc,
	  64, 64, 64-x, y );

} /* End of Update_Magniphase() */

/*------------------------------------------------------------------------*/

/*  Set_TxRx_Labels()
 *
 *  Sets up the labels in the Tx/Rx frames
 */

  void
Set_TxRx_Labels( void )
{
  if( isFlagSet(TRANSMIT_KEYBD) ||
	  isFlagSet(TRANSMIT_MACRO) )
  {
	if( isFlagSet(MODE_BPSK) )
	  gtk_label_set_markup( GTK_LABEL(lookup_widget(main_window, "xmit_status")),
		  XMIT_BPSK );
	else
	  if( isFlagSet(MODE_QPSK_USB) )
		gtk_label_set_markup( GTK_LABEL(lookup_widget(main_window, "xmit_status")),
			XMIT_QPSK_USB );
	  else
		if( isFlagSet(MODE_QPSK_LSB) )
		  gtk_label_set_markup( GTK_LABEL(lookup_widget(main_window, "xmit_status")),
			  XMIT_QPSK_LSB );
  }
  else
	gtk_label_set_markup( GTK_LABEL(lookup_widget(main_window, "xmit_status")),
		XMIT_OFF );

  if( isFlagSet(RECEIVE_MODE) )
  {
	if( isFlagSet(MODE_BPSK) )
	  gtk_label_set_markup( GTK_LABEL(lookup_widget(main_window, "rcve_status")),
		  RECV_BPSK );
	else
	  if( isFlagSet(MODE_QPSK_USB) )
		gtk_label_set_markup( GTK_LABEL(lookup_widget(main_window, "rcve_status")),
			RECV_QPSK_USB );
	  else
		if( isFlagSet(MODE_QPSK_LSB) )
		  gtk_label_set_markup( GTK_LABEL(lookup_widget(main_window, "rcve_status")),
			  RECV_QPSK_LSB );
  }
  else
	gtk_label_set_markup( GTK_LABEL(lookup_widget(main_window, "rcve_status")),
		RECV_OFF );

} /* Set_TxRx_Labels() */

/*------------------------------------------------------------------------*/

/*  Clear_Record_Fields()
 *
 *  Clears all QSO Record fields
 */

  void
Clear_Record_Fields( gboolean all )
{
  gtk_entry_set_text( GTK_ENTRY(lookup_widget(main_window, "callsign")), "" );
  gtk_entry_set_text( GTK_ENTRY(lookup_widget(main_window, "rst_in")), "" );
  gtk_entry_set_text( GTK_ENTRY(lookup_widget(main_window, "rst_out")), "" );
  gtk_entry_set_text( GTK_ENTRY(lookup_widget(main_window, "op_name")), "" );
  gtk_entry_set_text( GTK_ENTRY(lookup_widget(main_window, "qth_name")), "" );
  gtk_entry_set_text( GTK_ENTRY(lookup_widget(main_window, "locator")), "" );
  if( all )
	gtk_entry_set_text( GTK_ENTRY(lookup_widget(main_window, "band")), "" );

  qso_record.dx_call[0] = '\0';
  qso_record.my_rst[0]  = '\0';
  qso_record.dx_rst[0]  = '\0';
  qso_record.dx_qth[0]  = '\0';
  qso_record.dx_loc[0]  = '\0';
  if( all )
	qso_record.freq[0]  = '\0';
  strcpy( qso_record.dx_name, "OP" );

  ClearFlag( SAVE_RECORD );
}

/*------------------------------------------------------------------------*/

/*  Print_Character()
 *
 *  Prints a character to a text view port
 */

void
Print_Character(
	int printchr,
	GtkTextBuffer *text_buffer,
	GtkWidget     *scrolledwindow,
	GtkTextIter   *iter )
{
  GtkAdjustment *adjustment;
  char char2text[2];

  /* Delete a char on backspace */
  if( printchr == BS )
  {
	GtkTextIter start;

	gtk_text_buffer_get_iter_at_offset( text_buffer, &start,
		gtk_text_buffer_get_char_count
		(text_buffer) -1 );
	gtk_text_buffer_delete( text_buffer, &start, iter );
	return;
  }

  /* Convert char to "text" */
  char2text[0] = (char) printchr;
  char2text[1] = '\0';

  /* Print character */
  gtk_text_buffer_insert( text_buffer, iter,
	  char2text, -1 );

  /* Record to QSO record file */
  if( isFlagSet(RECORD_QSO) )
	fprintf( qso_record_fp, "%c", printchr );

  /* Scroll Text View to bottom on LF */
  if( (printchr == LF) ||
	  (printchr == GDK_Return) )
  {
	adjustment = gtk_scrolled_window_get_vadjustment
	  ( GTK_SCROLLED_WINDOW(scrolledwindow) );
	gtk_adjustment_set_value( adjustment, adjustment->upper );

	if( isFlagSet(RECEIVE_MODE) )
	  fprintf( qso_record_fp, "Rx-> " );
	else
	  fprintf( qso_record_fp, "Tx-> " );
  }

} /* Print_Character() */

/*------------------------------------------------------------------------*/

