/* operation.c
 *
 *  Operating modes functions for xpsk31
 */

/*
 *  xpsk31: An application to transmit and receive
 *  PSK31 signals using a computer's sound card
 *
 *  Copyright (C) 2001  Neoklis Kyriazis
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details:
 *
 *  http://www.gnu.org/copyleft/gpl.txt
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gtk/gtk.h>
#include "xpsk31.h"
#include "support.h"

/* xpsk31 main window */
extern GtkWidget *main_window;

/* Global widgets */
extern GtkWidget
  *gbl_rx_scrolledwindow,
  *gbl_tx_scrolledwindow;

/* Text buffers for text viewers */
extern GtkTextBuffer
  *gbl_rx_text_buffer,
  *gbl_tx_text_buffer;

/* Keyboard entry buffer */
extern guint key_buff[32];
/* Keystrokes counter */
extern int key_cnt;

extern FILE
  /* File for recording QSO's */
  *qso_record_fp,
  /* File for ADIF QSO record */
  *log_adif_fp,
  /* File for station log */
  *log_fp,
  /* Pointer to samples file */
  *samples_fp;

extern char
  label[9][17], /* Labels for F1-F9 commands */
  *macro[9];    /* Macros attached to F1-F9  */

/* Index to macros */
int gbl_macro_idx;

/* QSO data record */
qso_record_t qso_record;

/* Runtime config data */
extern rc_data_t rc_data;

/*------------------------------------------------------------------------*/

/*  Transmit_Keybd()
 *
 *  Transmits directly from the keyboard
 */

  gboolean
Transmit_Keybd( gpointer data )
{
  /* Text buffer marker */
  static GtkTextIter iter;

  /* First call of function flag */
  static gboolean first_call = TRUE;

  /* Count characters entered at keyboard */
  static int char_cnt = 0;

  /* Pointer to current char in keybd buffer */
  static int key_pnt = 0;

  /*** Cancel Transmit if Tx flag cleared ***/
  if( isFlagClear(TRANSMIT_KEYBD) )
  {
	/* Switch to CW(N) and PTT OFF */
	if( isFlagClear(TRANSMIT_MACRO) )
	{
	  Transmit_Postamble();
	  if( isFlagSet(DSP_IO_ERROR) )
		return( FALSE );
	  ClearFlag( TRANSMIT_MODE );
	}

	first_call = TRUE;
	return( FALSE );

  } /* if( isFlagClear(TRANSMIT_KEYBD) ) */

  /*** Initialize if needed ***/
  if( first_call )
  {
	first_call = FALSE;
	gtk_widget_grab_focus( lookup_widget(main_window, "tx_textview") );
	if( isFlagClear(TRANSMIT_MODE) )
	  if( !Initialize_Transmit(data) )
	  {
		first_call = TRUE;
		return( FALSE );
	  }

	return( TRUE );
  } /* if( first_call ) */

  /*** Process keystrokes from user ***/
  if( key_cnt != key_pnt )
  {
	/* Set up TRANSMIT text view */
	gtk_text_buffer_get_iter_at_offset(
		gbl_tx_text_buffer, &iter,
		gtk_text_buffer_get_char_count
		(gbl_tx_text_buffer) );

	/* Capitalize letters if enabled */
	if( isFlagSet(CAPITAL_LETTERS) &&
		(key_buff[key_pnt] > 0x60) &&
		(key_buff[key_pnt] < 0x7b) )
	  key_buff[key_pnt] -= 0x20;

	/* Send LF and CR on Return */
	if( key_buff[key_pnt] == GDK_Return )
	{
	  char_cnt = 0; /* Reset word wrap */
	  Transmit_PSK_Character( CR );
	  Transmit_PSK_Character( LF );
	  Print_Character( GDK_Return, gbl_tx_text_buffer,
		  gbl_tx_scrolledwindow, &iter );

	  /* Reset key buff pointer */
	  if( ++key_pnt == 32 )
		key_pnt = 0;

	  return( TRUE );
	}

	/* Do word wrapping */
	if( (char_cnt >= rc_data.wwrap) &&
		(key_buff[key_pnt] == ' ') )
	{
	  char_cnt = 0; /* Reset word wrap */
	  Transmit_PSK_Character( CR );
	  Transmit_PSK_Character( LF );
	  Print_Character( GDK_Return, gbl_tx_text_buffer,
		  gbl_tx_scrolledwindow, &iter );

	  /* Reset key buff pointer */
	  if( ++key_pnt == 32 )
		key_pnt = 0;

	  return( TRUE );
	}

	/* Backspace cursor */
	if( key_buff[key_pnt] == GDK_BackSpace )
	{
	  char_cnt--;
	  key_buff[key_pnt] = BS;
	}

	/* Transmit keyboard char and record */
	char_cnt++;
	Transmit_PSK_Character( key_buff[key_pnt] );
	Print_Character( key_buff[key_pnt], gbl_tx_text_buffer,
		gbl_tx_scrolledwindow, &iter );

	/* Reset key buff pointer */
	if( ++key_pnt == 32 )
	  key_pnt = 0;
  }
  else
	/* Transmit 'null' character in idle periods */
	Transmit_PSK_Character( NULL_CHAR );

  return( TRUE );

} /* End of Transmit_Keybd() */

/*------------------------------------------------------------------------*/

/*  Transmit_Macro()
 *
 *  Transmits a pre-recorded Macro
 */

  gboolean
Transmit_Macro( gpointer data )
{
  /* Text buffer marker */
  static GtkTextIter iter;

  static int
	macro_cnt = 0, /* Count macro characters tx'd */
    tag_idx   = 0; /* Index for transmitting tags */

  /* First call of function flag */
  static gboolean first_call = TRUE;

  /* Tags for entering fields in macros */
  char *tags[NUM_OF_TAGS] = TAG_TABLE;

  /* String within a tag <> */
  char tag_str[10];

  /* Tag replacement string */
  static char tag_rep[22];


  /* Abort macro if user disabled, remain in Tx mode */
  /* Stop macro on '~' character, remain in Tx mode  */
  if( isFlagClear(TRANSMIT_MACRO) ||
	  (macro[gbl_macro_idx][macro_cnt] == '~') )
  {
	Transmit_PSK_Character( NULL_CHAR );
	fflush( qso_record_fp );
	macro_cnt = tag_idx = 0;
	first_call = TRUE;
	ClearFlag( TRANSMIT_MACRO |	TRANSMIT_TAG );
	SetFlag( TRANSMIT_KEYBD );
	gtk_idle_add( Transmit_Keybd, "from_macro" );
	return( FALSE );
  }

  /* Stop macro at string terminator, */
  /* or CW id char (%), go to Rx mode */
  if( (macro[gbl_macro_idx][macro_cnt] == '\0') ||
	  (macro[gbl_macro_idx][macro_cnt] == '*') )
  {
	Transmit_PSK_Character( NULL_CHAR );
	Transmit_Postamble();
	if( isFlagSet(DSP_IO_ERROR) )
	  return( FALSE );
	fflush( qso_record_fp );

	/* Transmit Morse ID */
	if( macro[gbl_macro_idx][macro_cnt] == '*' )
	{
	  char message[24];

	  snprintf( message, 24, "  DE %s  TU", rc_data.call );
	  Morse_Transmit( message );
	  if( isFlagSet(DSP_IO_ERROR) )
		return( FALSE );
	}

	macro_cnt = tag_idx = 0;
	first_call = TRUE;
	ClearFlag(
		TRANSMIT_MACRO |
		TRANSMIT_KEYBD |
		TRANSMIT_MODE  |
		TRANSMIT_TAG );
	SetFlag( RECEIVE_MODE );
	gtk_idle_add( Receive_Mode, NULL );
	return( FALSE );
  }

  /*** Initialize if needed ***/
  if( first_call )
  {
	first_call = FALSE;
	if( isFlagClear(TRANSMIT_MODE) )
	  if( !Initialize_Transmit(data) )
	  {
		first_call = TRUE;
		return( FALSE );
	  }

	return( TRUE );
  } /* if( first_call ) */

  /* Set up TRANSMIT text view */
  gtk_text_buffer_get_iter_at_offset( gbl_tx_text_buffer, &iter,
	  gtk_text_buffer_get_char_count
	  (gbl_tx_text_buffer) );

  /* If Macro begins with a ^ transmit new line */
  if( macro[gbl_macro_idx][macro_cnt] == '^' )
  {
	Transmit_PSK_Character( CR );
	Transmit_PSK_Character( LF );
	Print_Character( LF, gbl_tx_text_buffer,
		gbl_tx_scrolledwindow, &iter );
	macro_cnt++;
	return( TRUE );
  }

  /*** Look for tags and substitude accordingly ***/
  if( (macro[gbl_macro_idx][macro_cnt] == '<') &&
	  isFlagClear(TRANSMIT_TAG) )
  {
	int idx;

	SetFlag( TRANSMIT_TAG );

	/*** Copy tag string to buffer and identify ***/
	macro_cnt++; /* Move into tag field */

	/* Copy line to tag, max 12 chars or EOS */
	idx = 0;
	while( (idx < 12) &&
		(macro[gbl_macro_idx][macro_cnt] != '>') )
	  tag_str[idx++] = macro[gbl_macro_idx][macro_cnt++];

	/* Terminate tag buffer and advance macro buffer */
	tag_str[idx] = '\0';
	macro_cnt++;

	/* Identify tag and substitude accordingly */
	idx = 0;
	while( idx < NUM_OF_TAGS )
	{
	  if( strcmp(tag_str, tags[idx]) == 0 )
		break;
	  idx++;
	}

	/* Abort if tag unidentified */
	if( idx == NUM_OF_TAGS )
	{
	  Error_Dialog(
		  "Error reading xpsk31rc file\n"
		  "A tag in a macro is invalid\n"
		  "Quit xpsk31 and correct", TRUE);
	  return( FALSE );
	}

	/* Replace tags */
	switch( idx )
	{
	  case 0: /* own-call */
		strncpy( tag_rep, rc_data.call, 17 );
		break;

	  case 1: /* own-name */
		strncpy( tag_rep, rc_data.name, 17 );
		break;

	  case 2: /* own-qth */
		strncpy( tag_rep, rc_data.qth, 21 );
		break;

	  case 3: /* own-loc */
		strncpy( tag_rep, rc_data.loc, 7 );
		break;

	  case 4: /* own-rst */
		strncpy( tag_rep, qso_record.my_rst, 4 );
		break;

	  case 5: /* rem-call */
		strncpy( tag_rep, qso_record.dx_call, 15 );
		break;

	  case 6: /* rem-name */
		strncpy( tag_rep, qso_record.dx_name, 13 );
		break;

	  case 7: /* rem-qth */
		strncpy( tag_rep, qso_record.dx_qth, 13 );
		break;

	  case 8: /* rem-loc */
		strncpy( tag_rep, qso_record.dx_loc, 7 );
		break;

	  case 9: /* rem-rst */
		strncpy( tag_rep, qso_record.dx_rst, 4 );
		break;

	  case 10: /* date-time */
		strncpy( tag_rep, qso_record.date, 12 );
		strcat ( tag_rep, " " );
		strncat( tag_rep, qso_record.time, 9 );
		break;

	  case 11: /* op-freq, strip trailing spaces */
		{
		  int i;

		  strncpy( tag_rep, qso_record.freq, 14 );
		  i = strlen( tag_rep ) - 1;
		  for( ; i >= 0; i-- )
			if( tag_rep[i] == ' ' )
			  tag_rep[i] = '\0';
		}
		break;

	  case 12: /* app-version */
		snprintf( tag_rep, 22, "%s %s", PACKAGE, VERSION );

	} /* switch( idx ) */

  } /* if( (macro[gbl_macro_idx][macro_cnt] == '<') && ) */

  /*** Transmit tag replacement ***/
  if( isFlagSet(TRANSMIT_TAG) )
  {
	if( tag_rep[tag_idx] != '\0' )
	{
	  /* Capitalize letters if enabled */
	  if( isFlagSet(CAPITAL_LETTERS) &&
		  (tag_rep[tag_idx] > 0x60)  &&
		  (tag_rep[tag_idx] < 0x7b) )
		tag_rep[tag_idx] -= 0x20;

	  Transmit_PSK_Character( tag_rep[tag_idx] );
	  Print_Character( tag_rep[tag_idx], gbl_tx_text_buffer,
		  gbl_tx_scrolledwindow, &iter );
	  tag_idx++;

	} /* if( (tag_rep[tag_idx] != '\0') ) */
	else
	{
	  ClearFlag( TRANSMIT_TAG );
	  tag_idx = 0;
	}

	return( TRUE );

  } /* if( isFlagSet(TRANSMIT_TAG) && */

  /* Capitalize letters if enabled */
  if( isFlagSet(CAPITAL_LETTERS)               &&
	  (macro[gbl_macro_idx][macro_cnt] > 0x60) &&
	  (macro[gbl_macro_idx][macro_cnt] < 0x7b) )
	macro[gbl_macro_idx][macro_cnt] -= 0x20;

  /* Transmit a char from Macro */
  Transmit_PSK_Character(macro[gbl_macro_idx][macro_cnt] );
  Print_Character( macro[gbl_macro_idx][macro_cnt],
	  gbl_tx_text_buffer, gbl_tx_scrolledwindow, &iter );
  macro_cnt++;

  return( TRUE );

} /* Transmit_Macro() */

/*------------------------------------------------------------------------*/

  gboolean
Receive_Mode( gpointer data )
{
  /* Text buffer marker */
  static GtkTextIter iter;

  /* First call of function flag */
  static gboolean first_call = TRUE;

  /* A character decoded from PSK31 signals */
  int dec_chr;

  /* Received char count */
  static int chr_cnt = 0;


  /* Cancel Receive if Rx flag cleared */
  if( isFlagClear(RECEIVE_MODE) )
  {
	first_call = TRUE;
	return( FALSE );
  }

  /*** Initialize ***/
  if( first_call )
  {
	first_call = FALSE;

	/* Setup sound card, abort on error */
	Setup_Sound_Card();
	if( isFlagClear(SOUND_CARD_SETUP) )
	{
	  ClearFlag( RECEIVE_MODE );
	  Set_TxRx_Labels();
	  return( FALSE );
	}

	/* Set transceiver mode if CAT enabled, abort on error */
	if( isFlagSet(ENABLE_CAT) )
	{
	  Set_Tcvr_Mode();
	  if( isFlagSet(SERIAL_IO_ERROR) )
	  {
		ClearFlag( RECEIVE_MODE );
		Set_TxRx_Labels();
		first_call = TRUE;
		return( FALSE );
	  }
	}

	/* Abort if QSO record file fails to open */
	if( !Open_Record_File() )
	  return( FALSE );

	/* Print new line to Rx viewer */
	gtk_text_buffer_get_iter_at_offset(
		gbl_rx_text_buffer, &iter,
		gtk_text_buffer_get_char_count(
		  gbl_rx_text_buffer) );

	Print_Character(
		LF, gbl_rx_text_buffer,
		gbl_rx_scrolledwindow, &iter );

	/* Reset char count on line feed */
	chr_cnt = 0;

	Set_TxRx_Labels();

	return( TRUE );

  } /* if( first_call ) */

  /*** Print characters to Receive window & file ***/
  /* Set up RECEIVE text view */
  gtk_text_buffer_get_iter_at_offset(
	  gbl_rx_text_buffer, &iter,
	  gtk_text_buffer_get_char_count
	  (gbl_rx_text_buffer) );

  /* Read a character, abort on dsp error */
  dec_chr = Decode_PSK_Character();
  if( isFlagSet(DSP_IO_ERROR) )
	return( FALSE );

  if( dec_chr <= 127 )
  {
	/* Capitalize letters if enabled */
	if( isFlagSet(CAPITAL_LETTERS) &&
		(dec_chr > 0x60)           &&
		(dec_chr < 0x7b) )
	  dec_chr -= 0x20;

	/* Print character to Rx viewer */
	Print_Character( dec_chr, gbl_rx_text_buffer,
		gbl_rx_scrolledwindow, &iter );

	/* Reset char count on line feed */
	if( dec_chr == LF )
	  chr_cnt = 0;
	else
	  /* Decrease char count on backspace */
	  if( dec_chr == BS )
		chr_cnt--;
	  else
	  {
		/* Do word wrapping */
		chr_cnt++;
		if( (chr_cnt >= rc_data.wwrap) &&
			(dec_chr == ' ') )
		{
		  Print_Character( LF, gbl_rx_text_buffer,
			  gbl_rx_scrolledwindow, &iter );
		  chr_cnt = 0;
		}
	  } /* else */

  } /* if( dec_chr <= 127 ) */

  return( TRUE );

} /* End of Receive_Mode() */

/*------------------------------------------------------------------------*/

/*  Initialize_Transmit()
 *
 *  Initializes the Transmit mode
 */

  gboolean
Initialize_Transmit(gpointer data)
{
  /* Setup sound card, abort on error */
  Setup_Sound_Card();
  if( isFlagClear(SOUND_CARD_SETUP) )
  {
	ClearFlag(
		TRANSMIT_KEYBD |
		TRANSMIT_MACRO |
		TRANSMIT_MODE );
	Set_TxRx_Labels();
	return( FALSE );
  }

  /* Set transceiver mode if CAT enabled, abort on error */
  if( isFlagSet(ENABLE_CAT) )
  {
	/* Switch to SSB and PTT ON */
	SetFlag( TRANSMIT_MODE );
	Set_Tcvr_Mode();
	if( isFlagSet(SERIAL_IO_ERROR) )
	{
	  ClearFlag( TRANSMIT_MODE );
	  Set_TxRx_Labels();
	  return( FALSE );
	}

  } /* if( isFlagSet(ENABLE_CAT) ) */

  Set_TxRx_Labels();

  /* Transmit carrier and reversals, abort on error */
  if( strcmp(data, "from_macro") != 0 )
  {
	Transmit_Preamble();
	if( isFlagSet(DSP_IO_ERROR) )
	  return( FALSE );
  }

  return( Open_Record_File() );

} /* Initialize_Transmit() */

/*------------------------------------------------------------------------*/

/*  Open_Record_File()
 *
 *  Opens a file for raw QSO recording
 */

  gboolean
Open_Record_File(void)
{
  /* File path for QSO record  */
  char qso_record_fpath[64];


  /* Open file for QSO Record, abort on error */
  if( isFlagSet(RECORD_QSO) )
  {
	/* Create a file for recording QSO's */
	if( qso_record_fp == NULL )
	{
	  /* Make file path for record file */
	  snprintf( qso_record_fpath, 64,
		  "%s/xpsk31/record.txt", getenv("HOME") );

	  qso_record_fp = fopen( qso_record_fpath, "a" );
	  if( qso_record_fp == NULL )
	  {
		perror( qso_record_fpath );
		Error_Dialog(
			"Failed to open QSO Record file\n"
			"Quit xpsk31 and correct", TRUE );
		return( FALSE );
	  }

	} /* if( QSO_record_fp == NULL ) */

  } /* if( isFlagSet(RECORD_QSO) ) */

  return( TRUE );

} /* Open_Record_File(void) */

/*------------------------------------------------------------------------*/

